-- Author: Ashnal
-- BoostLib is a library that other scripts can use to define boosts
-- using it ensures that the new boosts will work nicely with Boost Stacking and boost chaining,
-- and it doesnt require hacks like re-using the MT or sneakertimer in most cases

if BLib then return end

local boostreadout
local cvars = {
    blib_boostreadout = {
        flags = CV_CALL,
        defaultvalue = "Off",
        PossibleValue = CV_OnOff,
        func = function(cvar)
            boostreadout = cvar.value
        end
    },
    blib_booststack = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = "On",
        PossibleValue = CV_OnOff,
        func = onlyAdjustableOnLevelChange
    },
    blib_boostchain = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = "On",
        PossibleValue = CV_OnOff,
        func = onlyAdjustableOnLevelChange
    },
    slipstream_boostchain = { -- This is here for Acro 0.5a compat, which looks for this cvar by name
        defaultvalue = "On",
        PossibleValue = CV_OnOff
    },
    blib_diminishingreturns = {
        flags = CV_NETVAR|CV_NOSHOWHELP|CV_CHEAT|CV_CALL,
        defaultvalue = "On",
        PossibleValue = CV_OnOff,
        func = onlyAdjustableOnLevelChange
    },
    blib_diminishingreturnsamount = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "1.60",
        PossibleValue = {MIN = FRACUNIT+FRACUNIT/4, MAX = FRACUNIT*2+FRACUNIT/3},
        func = onlyAdjustableOnLevelChange
    },
	blib_airdrag = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = "On",
        PossibleValue = CV_OnOff,
        func = onlyAdjustableOnLevelChange
    },
    blib_debugboost = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "0",
        PossibleValue = {MIN = 0, MAX = INT32_MAX},
        func = onlyAdjustableOnLevelChange
    },
    blib_debugaccel = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "0",
        PossibleValue = {MIN = 0, MAX = INT32_MAX},
        func = onlyAdjustableOnLevelChange
    }
}

for k,v in pairs(cvars) do
	v.name = k
	v.ref = CV_RegisterVar(v) -- use a script scope local with CV_CALL to keep track of it, it performs better that way
	rawset(_G, k, v.ref)
end

local cv_legacyboostchain = cvars.slipstream_boostchain.ref -- because this var's needs are odd

local booststack
local boostchain
local diminish
local diminishamount
local airdrag
local debugboost
local debugaccel
addHook("MapChange", function(gamemap)
    booststack = blib_booststack.value 
    boostchain = blib_boostchain.value 
    diminish = blib_diminishingreturns.value 
	diminishamount = blib_diminishingreturnsamount.value
	airdrag = blib_airdrag.value
	debugboost = blib_debugboost.value
	debugaccel = blib_debugaccel.value
end)
addHook("NetVars", function(net)
    booststack = net($)
    boostchain = net($)
    diminish = net($) 
	diminishamount = net($)
	airdrag = net($)
	debugboost = net($)
	debugaccel = net($)
end)

--fucking local constants to make everything perform better, because srb2 does a string lookup for hardcode constants, which tanks performance
local FRACUNIT = FRACUNIT
local FRACBITS = FRACBITS

--kartstuff constants
local k_sneakertimer = k_sneakertimer
local k_spinouttimer = k_spinouttimer
local k_wipeoutslow = k_wipeoutslow
local k_drift = k_drift
local k_speedboost = k_speedboost
local k_accelboost = k_accelboost
local k_aizdriftstrat = k_aizdriftstrat
local k_offroad = k_offroad
local k_driftboost = k_driftboost
local k_boostpower = k_boostpower
local sfx_cdfm70 = sfx_cdfm70
local MFE_SPRUNG = MFE_SPRUNG

local FixedMul = FixedMul
local FixedDiv = FixedDiv
local min = min
local max = max
local P_IsObjectOnGround = P_IsObjectOnGround
local pairs = pairs
local P_ReturnThrustX = P_ReturnThrustX
local P_ReturnThrustY = P_ReturnThrustY
local S_SoundPlaying = S_SoundPlaying
local P_InstaThrust = P_InstaThrust
local R_PointToAngle2 = R_PointToAngle2
local FixedHypot = FixedHypot
local pairs = pairs

local BT_BRAKE = BT_BRAKE

local pmo,pks

local offroadaffected

freeslot("SPR_BSSS")
freeslot("S_STACKINGVFX1")
local S_STACKINGVFX1 = S_STACKINGVFX1
states[S_STACKINGVFX1] = {SPR_BSSS, FF_TRANS30|FF_ANIMATE|A, -1, nil, 4, 2, S_STACKINGVFX1}

local function SpawnStackVFX(mo, color)
	local stack = P_SpawnMobj(mo.x, mo.y, mo.z, MT_THOK)
	stack.state = S_STACKINGVFX1
	stack.flags = MF_NOCLIP|MF_NOCLIPHEIGHT|MF_NOGRAVITY|MF_DONTENCOREMAP|MF_NOBLOCKMAP
	stack.radius = 24*FRACUNIT
	stack.height = 100*FRACUNIT
	stack.target = mo
	stack.color = color
	stack.colorized = true
	stack.stackingvfx = true
	return stack
end

-- prefer to input all fixed values
-- https://en.wikipedia.org/wiki/Feature_scaling
local function Rescale(value, oldmin, oldmax, newmin, newmax)
	return newmin +  FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function Clamp(value, newmin ,newmax)
	return min(newmax,max(newmin,value))
end

local function FixedSpeedToFloatString(fixedspeed)
	local scaledspeed = FixedDiv(fixedspeed, mapobjectscale)
	local frac = ((scaledspeed & UINT16_MAX)*100)>>FRACBITS
	frac = frac < 10 and "0"..frac or frac
	return scaledspeed>>FRACBITS.."."..frac
end

local travelangle
local newx,newy
addHook("MobjThinker", function(mo)
	if not mo.stackingvfx then return end
	if not (mo.target and mo.target.valid and mo.target.player and mo.target.player.STACKINGVFX and not mo.player) then
		P_RemoveMobj(mo)
	else
		travelangle = mo.target.angle-(ANG1*9)*mo.target.player.kartstuff[k_drift]
		newx = mo.target.x + P_ReturnThrustX(mo.target, travelangle, FixedMul(1310720, mo.scale)) -- 20*FRACUNIT
		newy = mo.target.y + P_ReturnThrustY(mo.target, travelangle, FixedMul(1310720, mo.scale)) -- 20*FRACUNIT
		mo.destscale = FixedMul(mo.target.scale, Rescale(mo.target.player.vfxstackedboosts*FRACUNIT, 2*FRACUNIT, 4*FRACUNIT,  FRACUNIT, 2*FRACUNIT))
		P_MoveOrigin(mo, newx, newy, mo.target.z - (FixedMul(mo.info.height, mo.scale) - FixedMul(mo.info.height,mo.target.scale))/2)
	end
end, MT_THOK)

--lets just reimplement boosting lololololol
--by callmore#4947 and Ashnal#4043
local slopeboost
hud.add(function(v, p, c)
	if not boostreadout then return end
    if p.spectator then return end
	if p.BLcustomboosts and p.BLcustomboosts["slope"] then slopeboost = FixedMul(p.BLcustomboosts["slope"].speedboost, 100*FRACUNIT)>>FRACBITS else slopeboost = 0 end
	v.drawString(12,15, "    offi: " .. (p.offroadimmune and 1 or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,25, "    slip: " .. (p.handlingboost and 1 or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,35, "    trkt: " .. (p.trickboost or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,45, "    dftt: " .. p.kartstuff[k_driftboost],V_SNAPTOLEFT,"left")
	v.drawString(12,55, "    snkt: " .. p.kartstuff[k_sneakertimer],V_SNAPTOLEFT,"left")
	v.drawString(12,65, "nisnkt: " .. (p.mo and p.mo.naturalitemsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,75, "  isnkt: " .. (p.mo and p.mo.itemsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,85, "npsnkt: " .. (p.mo and p.mo.naturalpanelsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,95, "  psnkt: " .. (p.mo and p.mo.panelsneakertimer or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,105, "      bst: " .. FixedMul(p.kartstuff[k_boostpower], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
	v.drawString(12,115, "    sspd: " .. (slopeboost or 0),V_SNAPTOLEFT,"left")
	v.drawString(12,125, "    tspd: " .. FixedMul(p.totalboost, 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
    v.drawString(12,135, "    dspd: " .. FixedMul(p.kartstuff[k_speedboost], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
	v.drawString(12,145, "      acc: " .. FixedMul(p.kartstuff[k_accelboost], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
end)

local BLib = {}
local psscb
local chainfuncs = {}

BLib.RegisterChainFunc = function(name, chainfunc)
	chainfuncs[name] = chainfunc
end

BLib.SetBoost = function(player, name, speedboost, accelboost, boostmult, stack, stackvfx, chain, offroadimmune, handlingboost)
	if not (player and name) then return end
	if not player.BLcustomboosts[name] then player.BLcustomboosts[name] = {} end
	psscb = player.BLcustomboosts
	psscb[name].speedboost = speedboost or 0
	psscb[name].accelboost = accelboost or 0
	psscb[name].boostmult = boostmult or 0
	psscb[name].chain = chain or false
	psscb[name].offroadimmune = offroadimmune or false
	psscb[name].handlingboost = handlingboost or false
	-- Chainfunc registration is done on load
	if stack == nil then psscb[name].stack = true else psscb[name].stack = stack end
	if stackvfx == nil then psscb[name].stackvfx = true else psscb[name].stackvfx = stackvfx end
end

local highesttimer
BLib.BoostChain = function(player, name, timer)
	if not boostchain and player then return timer end
	highesttimer = 0
	for k,v in pairs(player.BLcustomboosts or {}) do
		if k == name or not v.chain then continue end -- Don't call your own chain function, or run chain functions for stuff that doesn't chain
		highesttimer = max(chainfuncs[k](player, name, timer) or 0, $) -- Note that this calls the chainfunc for each boost -- also, let it fail if no chainfunc is defined, so any plugin authors get aan error message
	end
	return highesttimer
end

rawset(_G, "BLib", BLib)

addHook("MobjThinker", function(mo)
	mo.BLonground = P_IsObjectOnGround(mo) -- I think doing this early and referencing it in all the other scripts will be faster?
end, MT_PLAYER)

local sneakerboost = { 53740+768, 32768, 17294+768}
local maxvanillaboost = { 53740+768, 32768, FRACUNIT*375/1000 }
local maxairspeedtarget = FRACUNIT*65
local pogoairspeedtarget = FRACUNIT*75
local scaledmaxairspeedtarget, newairspeedcap, scaledpogoairspeedtarget
local speedboost,accelboost,boostmult,speedmod,boostmod,intermediate
local nonstackboost, nonstackboostmult, hasChainableBoost, harddiminish, onground, diminishval
local handlingboost, offroadimmune, boostpower, braking
local function GetKartBoostPower(player, pmo)

	if pks[k_spinouttimer] and pks[k_wipeoutslow] == 1 then -- Slow down after you've been bumped
		pks[k_speedboost] = 0
		pks[k_accelboost] = 0
		return
	end

	speedboost = 0
	nonstackboost = 0
	accelboost = 0
	boostmult = FRACUNIT
	nonstackboostmult = FRACUNIT
	hasChainableBoost = false
	offroadimmune = false
	player.offroadimmune = false
	handlingboost = false
	player.handlingboost = false
	player.vfxstackedboosts = 0
	player.stackedboosts = 0

	if debugboost then
		speedboost = $ + debugboost
	end
	if debugaccel then
		accelboost = max($, debugaccel)
	end
	
	--print("- boosttable - "..player.name)
	for k,v in pairs(player.BLcustomboosts or {}) do			
		
		speedmod = v.speedboost -- Can't be nil
		boostmod = v.boostmult
	
		--print(k.." "..speedmod.." "..boostmod.." "..v.accelboost)
	
		if (speedmod or boostmod) and boostchain and v.chain then 
			hasChainableBoost = true 
		end
	
		if (speedmod or boostmod) and booststack then
			player.stackedboosts = $ + 1
			if v.stackvfx then
				player.vfxstackedboosts = $ + 1
			end
		end
		if v.stack and booststack then
			speedboost = $ + speedmod
			boostmult = $ + boostmod
		else
			nonstackboost = max($, speedmod)
			nonstackboostmult = max($, FRACUNIT + boostmod)
		end
		
		if v.handlingboost then 
			handlingboost = true
			player.handlingboost = true
		end
		if v.offroadimmune then 
			offroadimmune = true 
			player.offroadimmune = true
		end
		
		accelboost = max($, v.accelboost)
	end
	
	speedboost = max($, nonstackboost)
	
	if nonstackboostmult ~= FRACUNIT then -- we had a nonstacking boostmult mod -- this wierdness allows for negative stacking boostmult as long as no other nonstacking boostmult mods are active
		boostmult = max($, nonstackboostmult)
	end

	onground = pmo.BLonground
	offroadaffected = not offroadimmune and onground and pks[k_offroad]

	if player.vfxstackedboosts >= 2 and not offroadaffected then
		--SpawnFastLines(player, 5*FRACUNIT/3, player.skincolor+12)		
		if not player.STACKINGVFX and pmo then
			player.STACKINGVFX = SpawnStackVFX(pmo, player.skincolor)
		end
		pmo.dotrail = true
	else
		player.STACKINGVFX = nil
		pmo.dotrail = false
	end
	
	player.totalboost = speedboost -- saved for later display by boostreadout

	-- Vanilla boostpower calc redone with adjustments
	boostpower = FRACUNIT
	if pks[k_spinouttimer] and pks[k_wipeoutslow] == 1 then
		boostpower = 0
	else
		if pks[k_bananadrag] > TICRATE then
			boostpower = (4*boostpower)/5
		end
		if offroadaffected then
			boostpower = FixedDiv(boostpower, pks[k_offroad] + FRACUNIT)
		else
			boostpower = boostpower + FixedMul(speedboost, boostmult) - speedboost
		end
	end

	pks[k_boostpower] = boostpower

	-- WARNING: Math ahead
	-- Apply new diminishing returns equation -- OLD: https://www.desmos.com/calculator/eoxcakxvxy
	-- NEW: https://www.desmos.com/calculator/wnc8uw72m9
	if diminish then  -- two stage, linear unti 0.5, then we start falling off
		if gamespeed <= 1 and speedboost > 32768 then -- FRACUNIT/2 -- Normal, green line
			diminishval = boostchain and max(diminishamount - 8192, FRACUNIT) or diminishamount -- FRACUNIT/8
			intermediate = FixedDiv(FixedMul(diminishval,-32768) - 16384, -diminishamount+32768) -- FRACUNIT*-1/2 -- FRACUNIT/4
			speedboost = FixedMul(diminishamount,(FRACUNIT-FixedDiv(FRACUNIT,($+intermediate))))
		elseif gamespeed == 2 and speedboost > 24576 then -- FRACUNIT*375/1000 -- We hard speed, use the harsher curve, red line
			harddiminish = Rescale(diminishamount, FRACUNIT, 131072, 62259, 117964) -- 2*FRACUNIT -- FRACUNIT*95/100 -- FRACUNIT*180/100
			diminishval = boostchain and max(harddiminish - 8192, FRACUNIT) or harddiminish -- FRACUNIT/8
			intermediate = FixedDiv(FixedMul(diminishval,-40960) - 9216,-harddiminish+24576) -- FRACUNIT*-625/1000 -- FRACUNIT*375/1000
			speedboost = FixedMul(harddiminish,(FRACUNIT-FixedDiv(FRACUNIT,($+intermediate))))
		end
	end

	if offroadaffected then
		pks[k_sneakertimer] = 0
		pmo.dotrail = false
	elseif S_SoundPlaying(pmo, sfx_cdfm70) then -- not offroadaffected, and the offroad sound is playing, so stop it
		S_StopSoundByID(pmo, sfx_cdfm70)
	end

	if offroadaffected then -- We behave as if the player has no boosts while in offroad, and smooth down
		pks[k_speedboost] = max(player.SSlastframespeedboost - player.SSlastframespeedboost/8, min(speedboost, 9831)) -- FRACUNIT*15/100 -- Apply strong falloff when in offroad and limit boost to a bit less than MT
	elseif speedboost >= (player.SSlastframespeedboost or 0) then
		pks[k_speedboost] = speedboost
	else
		pks[k_speedboost] = max(player.SSlastframespeedboost - 728, 0) -- value smoothing
	end
	
	pks[k_accelboost] = accelboost
	
	--player.speed = min($, FixedMul(FRACUNIT*128, mapobjectscale)) -- Turn right bug starts occuring somewhere between 130 and 140 FU/T, so we hard limit speed to prevent that
	--if pks[k_pogospring] then player.speed = min($, FixedMul(FRACUNIT*75, mapobjectscale)) end
	
	-- Award stuff
	player.highestboost = max($ or 0, pks[k_speedboost])
	player.higheststackcount = max($ or 0, player.stackedboosts)
	if hasChainableBoost then
		player.currentchainduration = ($ or 0) + 1
		player.longestchain = max($ or 0, player.currentchainduration)
	else
		player.currentchainduration = 0
	end

	player.SSlastframespeedboost = pks[k_speedboost]
end

local speed, newspeed, vanillaspeedcap, airspeedtarget
local function AirDrag(player, pmo)
	if not onground and pmo.lastonground then -- We launched
		if pmo.eflags & MFE_SPRUNG then
			speed = FixedHypot(pmo.momx,pmo.momy)
		else
			speed = player.speed
		end
		--local launchtype = pks[k_pogospring] and "pogo " or ""
		--print(launchtype.."launch player.speed: "..FixedSpeedToFloatString(player.speed).." launch mom: "..FixedSpeedToFloatString(FixedHypot(pmo.momx,pmo.momy)))
		pmo.launchspeed = min(speed, pks[k_pogospring] and INT32_MAX or newairspeedcap)
		pmo.lastairspeed = min(speed, pks[k_pogospring] and INT32_MAX or newairspeedcap)
		pmo.launchspeedz = pmo.momz
	elseif not onground and pmo.lastairspeed ~= nil then
		speed = min(FixedHypot(pmo.momx,pmo.momy), pks[k_pogospring] and INT32_MAX or newairspeedcap)
		--local debugstring = ""
		--debugstring = $.."pre-adjustment player.speed: "..FixedSpeedToFloatString(player.speed).."  mom: "..FixedSpeedToFloatString(speed).."  target: "..FixedSpeedToFloatString(scaledmaxairspeedtarget).."  lastairspeed: "..FixedSpeedToFloatString(pmo.lastairspeed)
		pmo.timeoffground = ($ or 0) + 1
		if speed == vanillaspeedcap then
			--if speed ~= pmo.lastairspeed then debugstring = $.." changed to "..FixedSpeedToFloatString(pmo.lastairspeed) end
			speed = pmo.lastairspeed
		end
		airspeedtarget = pks[k_pogospring] and scaledpogoairspeedtarget or scaledmaxairspeedtarget
		newspeed = Clamp(speed - (speed - airspeedtarget)/70, airspeedtarget, speed)
		P_InstaThrust(pmo, R_PointToAngle2(0, 0, pmo.momx*3, pmo.momy*3), newspeed)
		--debugstring = $.."  newspeed: "..FixedSpeedToFloatString(newspeed)
		--print(debugstring)
		pmo.lastairspeed = newspeed
	end
	pmo.lastonground = onground
end

addHook("MobjThinker", function(mo)
	if mo.valid and not mo.blchecked and mo.target and mo.target.valid and mo.target.player and mo.target.player.valid and mo.target.player.offroadimmune then
		mo.fuse = 1 -- setting fuse to 1 makes the object get removed in the later part of the hardcode mobjthinker after this hook
		mo.blchecked = true
	end
end, MT_WIPEOUTTRAIL)

addHook("MobjThinker", function(mo)
	if mo.valid and not mo.blchecked and mo.target and mo.target.valid and mo.target.player and mo.target.player.valid and mo.target.player.offroadimmune then
		mo.fuse = 1 -- setting fuse to 1 makes the object get removed in the later part of the hardcode mobjthinker after this hook
		mo.blchecked = true
	end
end, MT_SNEAKERTRAIL)

addHook("MapChange", function(mapnum)
	-- stupid hack for Acro's detection
	if cv_legacyboostchain.value ~= boostchain then
		CV_StealthSet(cv_legacyboostchain, boostchain)
	end
	vanillaspeedcap = 75*( mapheaderinfo[mapnum].mobj_scale or FRACUNIT )
	newairspeedcap = 120*( mapheaderinfo[mapnum].mobj_scale or FRACUNIT )
	scaledmaxairspeedtarget = FixedMul(maxairspeedtarget, mapheaderinfo[mapnum].mobj_scale or FRACUNIT)
	scaledpogoairspeedtarget = FixedMul(pogoairspeedtarget, mapheaderinfo[mapnum].mobj_scale or FRACUNIT)
	for player in players.iterate do
		player.SSlastframespeedboost = 0
		player.longestchain = 0
		player.higheststackcount = 0
		--player.BLcustomboosts = {}
	end
end)
addHook("NetVars", function(net)
	-- stupid hack for Acro's detection
	if cv_legacyboostchain.value ~= boostchain then
		CV_StealthSet(cv_legacyboostchain, boostchain)
	end
	vanillaspeedcap = net($)
	newairspeedcap = net($)
	scaledmaxairspeedtarget = net($)
	scaledpogoairspeedtarget = net($)
end)

addHook("PlayerSpawn", function(p)
	p.SSlastframespeedboost = 0
	p.BLcustomboosts = {}
end)

addHook("ThinkFrame", do
	for p in players.iterate do	
		pmo = p.mo
		pks = p.kartstuff
		if pmo then
			GetKartBoostPower(p, pmo)
			if airdrag then AirDrag(p, pmo) end
		end
	end
end)

addHook("MapLoad", hostmodload)
addHook("NetVars", hostmodload) -- Joiners don't run MapLoad