-- Intended to be turned on on top of hard speed

-- this is a good example of how to optimize cvar access for performance, a local is very fast for lua to access
-- so instead of grabbing the .value from the cvar every tic of the thinker, we check it on MapChange once, and store it in a local for reference when performance is important, like during thinkers or thinkframe
-- do note you'll NEED to sync these locals in a NetVars hook, or you'll desync
-- theres a func below that ensures players can't desync the real value of the var from the local when they change it, and makes it behave like kartspeed

if not BLib then return end
local SetBoost = BLib.SetBoost

local prefix = "blib_"
local cv_expert = CV_RegisterVar({
    name = prefix.."expert",
    defaultvalue = "Off",
    flags = CV_NETVAR|CV_CALL,
    PossibleValue = CV_OnOff,
	func = onlyAdjustableOnLevelChange
})

local cv_expertspeed = CV_RegisterVar({
    name = prefix.."expertspeed",
    defaultvalue = "0.09",
    flags = CV_NETVAR|CV_FLOAT|CV_CALL,
    PossibleValue = CV_Unsigned,
	func = onlyAdjustableOnLevelChange
})

local cv_expertboost = CV_RegisterVar({
    name = prefix.."expertboost",
    defaultvalue = "0.07",
    flags = CV_NETVAR|CV_FLOAT|CV_CALL,
    PossibleValue = CV_Unsigned,
	func = onlyAdjustableOnLevelChange
})

local cv_expertaccel = CV_RegisterVar({
    name = prefix.."expertaccel",
    defaultvalue = "0.31",
    flags = CV_NETVAR|CV_FLOAT|CV_CALL,
    PossibleValue = CV_Unsigned,
	func = onlyAdjustableOnLevelChange
})

local expert, expertspeed, expertboost, expertaccel
addHook("MapChange", function(gamemap)
	expert = cv_expert.value
	expertspeed = cv_expertspeed.value
	expertboost = cv_expertboost.value
	expertaccel = cv_expertaccel.value
end)
addHook("NetVars", function(net)
	expert = net($)
	expertspeed = net($)
	expertboost = net($)
	expertaccel = net($)
end)

local function setspeed(param)
	if expert then
		if type(param) == "number" then -- mapchange or mapload
			for p in players.iterate do
				SetBoost(p, "expert", expertspeed, expertaccel, expertboost, true, false, false)
			end
		else -- playerspawn
			SetBoost(param, "expert", expertspeed, expertaccel, expertboost, true, false, false)
		end
	end
end
addHook("MapChange", setspeed)
addHook("MapLoad", setspeed)
addHook("PlayerSpawn", setspeed)

--hostmod scoreboard
local function hmaddmod(fug)
	table.insert(balancechanges, fug)
end

addHook("MapLoad", hostmodload)
addHook("NetVars", hostmodload) -- Joiners don't run MapLoad