-- Author: Ashnal
-- Adds slipstreaming aka drafting like the later Mario Kart games
-- Youll see small gray speed lines when you're charging a slipstream boost
-- Theres a subtle woosh sound when it activates, and the gray speed lines will become normal size and more frequent
-- You need to have another racer within the maxangle of your heading, and be within the maxdistiance range to charge it
-- the closer you are to them, the faster it will charge
-- KNOWN BUG: the fast lines have some wierd angles on OpenGL rendering. No idea why.

if not BLib then return end
local SetBoost = BLib.SetBoost

local cvars = {
    ss_enabled = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = "On",
        PossibleValue = CV_OnOff,
        func = onlyAdjustableOnLevelChange
    },
    ss_colorized = {
        flags = CV_CALL,
        defaultvalue = "On",
        PossibleValue = CV_OnOff,
        func = onlyAdjustableOnLevelChange
    },
    ss_maxdistance = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = 2400, -- enough room to charge slowly, and dodge
        PossibleValue = CV_Unsigned,
        func = onlyAdjustableOnLevelChange
    },
    ss_t2chargedist = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = 1400,
        PossibleValue = CV_Unsigned,
        func = onlyAdjustableOnLevelChange
    },
    ss_t3chargedist = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = 300, -- danger close, woudln't be able to dodge an aimed item
        PossibleValue = CV_Unsigned,
        func = onlyAdjustableOnLevelChange
    },
    ss_maxangle = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "10", --10 degrees
        PossibleValue = CV_Unsigned,
        func = onlyAdjustableOnLevelChange
    },
    -- This is measured in tics
    ss_chargetoboost = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = 5*TICRATE/2,
        PossibleValue = CV_Unsigned,
        func = onlyAdjustableOnLevelChange
    },
    ss_boosttimer = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = 2*TICRATE,
        PossibleValue = CV_Unsigned,
        func = onlyAdjustableOnLevelChange
    },
    ss_minimumspeed = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = 28,
        PossibleValue = CV_Unsigned,
        func = onlyAdjustableOnLevelChange
    },
    -- Boost youll get if you're a 1 kartspeed
    ss_maxspeedboostpercent = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "0.27",
        PossibleValue = {MIN = 0, MAX = INT32_MAX},
        func = onlyAdjustableOnLevelChange
    },
    -- Boost youll get if you're a 9 or higher kartspeed
    ss_minspeedboostpercent = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "0.17",
        PossibleValue = {MIN = 0, MAX = INT32_MAX},
        func = onlyAdjustableOnLevelChange
    },
    ss_accelboostpercent = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "0.50",
        PossibleValue = {MIN = 0, MAX = INT32_MAX},
        func = onlyAdjustableOnLevelChange
    },
    ss_maxdraftspeedboostpercent = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "0.12",
        PossibleValue = {MIN = 0, MAX = INT32_MAX},
        func = onlyAdjustableOnLevelChange
    },
    ss_mindraftspeedboostpercent = {
        flags = CV_NETVAR|CV_FLOAT|CV_CALL,
        defaultvalue = "0.01",
        PossibleValue = {MIN = 0, MAX = INT32_MAX},
        func = onlyAdjustableOnLevelChange
    },
    --ss_anglefinder = {
    --    defaultvalue = "Off",
    --    PossibleValue = CV_OnOff
    --},
    ss_alwaysactive = {
        flags = CV_NETVAR|CV_CALL,
        defaultvalue = "Off",
        PossibleValue = CV_OnOff,
        func = onlyAdjustableOnLevelChange
    }
}

for k,v in pairs(cvars) do
	v.name = k
	v.ref = CV_RegisterVar(v) -- use a script scope local with CV_CALL to keep track of it, it performs better that way
	rawset(_G, k, v.ref)
end

local slipstream 
local colorized 
local maxdistance, maxdistancedrift
local t2chargedist 
local t3chargedist 
local maxangle, fixedmaxangle, fixedmaxangledrift
local chargetoboost 
local boosttime 
local minimumspeed 
local maxspeedboostpercent 
local minspeedboostpercent 
local accelboostpercent 
local maxdraftspeedboostpercent 
local mindraftspeedboostpercent 
--local anglefinder 
local alwayson 
local localtimetravel
addHook("MapChange", function(gamemap)
    slipstream = ss_enabled.value 
    colorized = ss_colorized.value 
    maxdistance = FixedMul(ss_maxdistance.value<<FRACBITS, mapheaderinfo[gamemap].mobj_scale or FRACUNIT)
	maxdistancedrift = 2*maxdistance/5 
    t2chargedist = FixedMul(ss_t2chargedist.value<<FRACBITS, mapheaderinfo[gamemap].mobj_scale or FRACUNIT)
    t3chargedist = FixedMul(ss_t3chargedist.value<<FRACBITS, mapheaderinfo[gamemap].mobj_scale or FRACUNIT)
    maxangle = ss_maxangle.value
	fixedmaxangle = FixedAngle(maxangle) -- Now were just precalcing everything possible in that critical block
	fixedmaxangledrift = FixedAngle(maxangle)
    chargetoboost = ss_chargetoboost.value 
    boosttime = ss_boosttimer.value 
    minimumspeed = FixedMul(ss_minimumspeed.value<<FRACBITS, mapheaderinfo[gamemap].mobj_scale or FRACUNIT)
    maxspeedboostpercent = ss_maxspeedboostpercent.value 
    minspeedboostpercent = ss_minspeedboostpercent.value 
    accelboostpercent = ss_accelboostpercent.value 
    maxdraftspeedboostpercent = ss_maxdraftspeedboostpercent.value 
    mindraftspeedboostpercent = ss_mindraftspeedboostpercent.value 
    --anglefinder = ss_anglefinder.value 
    alwayson = ss_alwaysactive.value
end)
addHook("MapLoad", function()
	localtimetravel = timetravel and timetravel.isActive 
end)
addHook("NetVars", function(net)
    slipstream = net($) 
    colorized = net($) 
    maxdistance = net($) 
	maxdistancedrift = net($)
    t2chargedist = net($) 
    t3chargedist = net($) 
    maxangle = net($)
	fixedmaxangle = net($) -- Now were just precalcing everything possible in that critical block
	fixedmaxangledrift = net($)
    chargetoboost = net($) 
    boosttime = net($) 
    minimumspeed = net($) 
    maxspeedboostpercent = net($) 
    minspeedboostpercent = net($) 
    accelboostpercent = net($) 
    maxdraftspeedboostpercent = net($) 
    mindraftspeedboostpercent = net($) 
    --anglefinder = net($) 
    alwayson = net($)
	localtimetravel = net($)
end)

freeslot(
"MT_ANGLEFINDER",
"S_ANGLEFINDER"
)

states[S_ANGLEFINDER] = {SPR_THOK, FF_TRANS30, 8, nil, 0, 0, S_ANGLEFINDER}

mobjinfo[MT_ANGLEFINDER] = {
	spawnstate = S_ANGLEFINDER,
	flags = MF_NOGRAVITY|MF_DONTENCOREMAP|MF_NOBLOCKMAP,
	radius = 1*FRACUNIT,
	height = 1*FRACUNIT
}

local INT32_MAX = INT32_MAX
local TICRATE = TICRATE
local FRACUNIT = FRACUNIT
local ANGLE_135 = ANGLE_135
local ANGLE_45 = ANGLE_45
local FRACBITS = FRACBITS
local SKINCOLOR_WHITE = SKINCOLOR_WHITE
local MT_FASTLINE = MT_FASTLINE
local MT_PARTICLE = MT_PARTICLE
local S_KART_SPIN = S_KART_SPIN

local k_hyudorotimer = k_hyudorotimer
local k_spinouttimer = k_spinouttimer
local k_wipeoutslow = k_wipeoutslow
local k_squishedtimer = k_squishedtimer
local k_drift = k_drift
local k_offroad = k_offroad

local FixedMul = FixedMul
local FixedDiv = FixedDiv
local min = min
local max = max
local FixedAngle = FixedAngle
local P_ReturnThrustX = P_ReturnThrustX
local P_ReturnThrustY = P_ReturnThrustY
local S_SoundPlaying = S_SoundPlaying
local P_InstaThrust = P_InstaThrust
local R_PointToAngle2 = R_PointToAngle2
local FixedHypot = FixedHypot
local P_SpawnMobj = P_SpawnMobj

-- prefer to input all fixed values
-- https://en.wikipedia.org/wiki/Feature_scaling
local function Rescale(value, oldmin, oldmax, newmin, newmax)
	return newmin +  FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function Clamp(value, newmin ,newmax)
	return min(newmax,max(newmin,value))
end

local newangleoffset = 0
local newx, newy, seekangle, tempdist, driftmod
addHook("MobjThinker", function(mo)
	if not anglefinder or not mo.target then
		P_RemoveMobj(mo)
	else
		seekangle = mo.target.player.mo.angle-(131235115)*mo.target.player.kartstuff[k_drift]
		
		if mo.target.player.kartstuff[k_drift] then 
			tempdist = FixedMul(mo.distanceoffset, 2*FRACUNIT/5)
			newangleoffset = mo.angleoffset * FixedAngle(maxangle*2)
		else 
			tempdist = FixedMul(mo.distanceoffset, FRACUNIT)
			newangleoffset = mo.angleoffset * FixedAngle(maxangle)
		end

		newx = mo.target.x + P_ReturnThrustX(mo.target, seekangle+newangleoffset, FixedMul(tempdist, mo.target.scale))
		newy = mo.target.y + P_ReturnThrustY(mo.target, seekangle+newangleoffset, FixedMul(tempdist, mo.target.scale))
		P_TeleportMove(mo, newx, newy, mo.target.z)
	end
end, MT_ANGLEFINDER)

local closeleft,closeright,closecenter
local farleft, farright, farcenter
local function AngleFinder()
	if anglefinder and displayplayers[0].mo then
		if not displayplayers[0].anglefinder then
		
			closeleft = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			closecenter = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			closeright = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			farleft = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			farcenter = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			farright = P_SpawnMobj(displayplayers[0].mo.x, displayplayers[0].mo.y, displayplayers[0].mo.z, MT_ANGLEFINDER)
			
			displayplayers[0].anglefinder = true
		end
		closeleft.target = displayplayers[0].mo
		closecenter.target = displayplayers[0].mo
		closeright.target = displayplayers[0].mo
		farleft.target = displayplayers[0].mo
		farcenter.target = displayplayers[0].mo
		farright.target = displayplayers[0].mo
		closeleft.scale = FRACUNIT/4
		closecenter.scale = FRACUNIT/4
		closeright.scale = FRACUNIT/4
		farleft.scale = FRACUNIT/4
		farcenter.scale = FRACUNIT/4
		farright.scale = FRACUNIT/4
		closeleft.distanceoffset = t3chargedist*FRACUNIT
		closecenter.distanceoffset = t3chargedist*FRACUNIT
		closeright.distanceoffset = t3chargedist*FRACUNIT
		farleft.distanceoffset = maxdistance*FRACUNIT
		farcenter.distanceoffset = maxdistance*FRACUNIT
		farright.distanceoffset = maxdistance*FRACUNIT
		closeleft.angleoffset = -1
		closecenter.angleoffset = 0
		closeright.angleoffset = 1
		farleft.angleoffset = -1
		farcenter.angleoffset = 0
		farright.angleoffset = 1
		closeleft.travel = true
		closecenter.travel = true
		closeright.travel = true
		farleft.travel = true
		farcenter.travel = true
		farright.travel = true
	else
		if displayplayers[0] then
			displayplayers[0].anglefinder = false
		end
	end
end

local fast
local function SpawnFastLines(mo, scale, color)
	fast = P_SpawnMobj(mo.x + (P_RandomRange(-36,36) * mo.scale),
					   mo.y + (P_RandomRange(-36,36) * mo.scale),
					   mo.z + (mo.height/2) + (P_RandomRange(-20,20) * mo.scale),
					   MT_FASTLINE)
	fast.angle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	fast.momx = 3*mo.momx/4
	fast.momy = 3*mo.momy/4
	fast.momz = 3*mo.momz/4
	fast.color = color
	fast.colorized = true
	fast.scale = FixedMul(scale,mo.scale)
	K_MatchGenericExtraFlags(fast, mo)
end

local angle,parts,i,verticalflip
local function SSDust(mo)
	angle = mo.angle
	verticalflip = mo.eflags & MFE_VERTICALFLIP
	if leveltime%2 then
		for i = 0,1 do
			newx = mo.x + P_ReturnThrustX(mo, angle + ((i == 0) and -1 or 1)*ANGLE_135, FixedMul(22*FRACUNIT, mo.scale))			
			newy = mo.y + P_ReturnThrustY(mo, angle + ((i == 0) and -1 or 1)*ANGLE_135, FixedMul(22*FRACUNIT, mo.scale))
			parts = P_SpawnMobj(newx, newy, (verticalflip and (mo.ceilingz+6*FRACUNIT*mapobjectscale or -1) or mo.floorz-6*FRACUNIT*mapobjectscale), MT_PARTICLE)
			parts.target = mo
			parts.angle = angle - ((i == 0) and -1 or 1)*ANGLE_45
			parts.scale = mo.scale
			parts.destscale = mo.scale*3
			parts.scalespeed = mo.scale/6
			parts.momx = 3*mo.momx/5
			parts.momy = 3*mo.momy/5				
			parts.fuse = 5
		end
	end
end


local clampedkartspeed,skincolor,draftingboost,slipstreamboost,slipstreamaccel,slipstreamtarget,angletotarget,disttotarget,seekangle,charge
local p,pks,tmo

local TIMETRAVEL_SLIPSTREAMINTEROP_VERSION = 2
if timetravel and (timetravel.SLIPSTREAMINTEROP_VERSION or 0) < TIMETRAVEL_SLIPSTREAMINTEROP_VERSION then
	timetravel.SLIPSTREAMINTEROP_VERSION = TIMETRAVEL_SLIPSTREAMINTEROP_VERSION
end


local driftvalue
local boosttype = "slipstream" -- referring to string constants as locals is faster than using literals

addHook("MobjThinker", function(mo)
	if not slipstream then return end -- Has to be turned on

	p = mo.player
	pks = mo.player.kartstuff
	
	if mo.state == S_KART_SPIN or pks[k_squishedtimer] then -- no slipstreaming if you've bumped or spun out
		mo.slipstreamtimer = 0
		mo.slipstreamcharge = 0
		mo.slipstreamsoundtimer = false
		SetBoost(p, boosttype)
		return
	end

	if mo.slipstreamcharge == nil then
		mo.slipstreamcharge = 0
		mo.slipstreamsoundtimer = false
	end

	-- reset until we find a valid slipstreamtarget this frame
	slipstreamboost = 0
	slipstreamaccel = 0
	slipstreamtarget = nil

	-- must be moving decently on the ground
	if mo.BLonground and p.speed >= minimumspeed then
		driftvalue = pks[k_drift]
		seekangle = mo.angle-(131235115)*driftvalue -- The magic number is 11 degrees, I'm optimizing this way >_>
		if localtimetravel then -- check this outside of the loop, and decide which loop to use for better performance, instead of checking inside the loop
			for potentialtarget in players.iterate do          
				tmo = potentialtarget.mo
				if ( tmo and tmo.valid -- must have valid player mapobject
					and tmo ~= mo -- Can't slipstream off yourself
					and not potentialtarget.kartstuff[k_hyudorotimer] -- or ghosts
					and tmo.BLonground -- or airbourne karts
					and potentialtarget.speed >= minimumspeed -- or slowpokes
				) then
					if mo.timetravel.isTimeWarped ~= tmo.timetravel.isTimeWarped and tmo.linkedItem and tmo.linkedItem.valid then
						tmo = tmo.linkedItem
					end
					-- Calculating the angle at which we look for suitable draft targets based on drifting
					disttotarget = FixedHypot(mo.x - tmo.x, mo.y - tmo.y)
					if driftvalue then
						if disttotarget > maxdistancedrift then continue end -- max slipstream distance while drifting
						if abs(seekangle - R_PointToAngle2(mo.x, mo.y, tmo.x, tmo.y)) > fixedmaxangledrift then continue end -- max angle check while drifting
					else 
						if disttotarget > maxdistance then continue end -- max slipstream distance
						if abs(seekangle - R_PointToAngle2(mo.x, mo.y, tmo.x, tmo.y)) > fixedmaxangle then continue end -- max angle check
					end -- Narrow angle for following to slipstream

					--print(mo.skin + " " + tmo.skin + " angletotarget: " + (abs(seekangle - R_PointToAngle2(mo.x, mo.y, tmo.x, tmo.y))) + " disttotarget: " + disttotarget)
					slipstreamtarget = potentialtarget
					break
				end
			end
		else
			for potentialtarget in players.iterate do          
				tmo = potentialtarget.mo
				if ( tmo and tmo.valid -- must have valid player mapobject
					and tmo ~= mo -- Can't slipstream off yourself
					and not potentialtarget.kartstuff[k_hyudorotimer] -- or ghosts
					and tmo.BLonground -- or airbourne karts
					and potentialtarget.speed >= minimumspeed -- or slowpokes
				) then				
					-- Calculating the angle at which we look for suitable draft targets based on drifting
					disttotarget = FixedHypot(mo.x - tmo.x, mo.y - tmo.y)
					if driftvalue then
						if disttotarget > maxdistancedrift then continue end -- max slipstream distance while drifting
						if abs(seekangle - R_PointToAngle2(mo.x, mo.y, tmo.x, tmo.y)) > fixedmaxangledrift then continue end -- max angle check while drifting
					else 
						if disttotarget > maxdistance then continue end -- max slipstream distance
						if abs(seekangle - R_PointToAngle2(mo.x, mo.y, tmo.x, tmo.y)) > fixedmaxangle then continue end -- max angle check
					end -- Narrow angle for following to slipstream

					--print(mo.skin + " " + tmo.skin + " angletotarget: " + angletotarget + " disttotarget: " + disttotarget)
					slipstreamtarget = potentialtarget
					break
				end
			end
		end
	end

	-- add/remove slipstream charge
	if slipstreamtarget then
		if disttotarget > t2chargedist then
			charge = 1
		elseif disttotarget > t3chargedist then
			charge = 2
		else
			charge = 2
		end

		mo.slipstreamcharge = min($+charge, chargetoboost)

		if mo.slipstreamcharge >= chargetoboost then
			mo.slipstreamtimer = boosttime
			if mo.slipstreamsoundtimer == false then
				S_StartSoundAtVolume(mo, sfx_s3k82, INT32_MAX)
				mo.slipstreamsoundtimer = true
			end
		end

		-- also spawn mini lines if charging to teach/show player the charging area
		if mo.slipstreamtimer == 0 then
			SpawnFastLines(mo, FRACUNIT/2, SKINCOLOR_WHITE)

			clampedkartspeed = Clamp(p.kartspeed, 1, 9)
			
			-- Set up drafting boost
			slipstreamboost = Rescale(clampedkartspeed*FRACUNIT, 9*FRACUNIT, FRACUNIT, mindraftspeedboostpercent, maxdraftspeedboostpercent)
			slipstreamaccel = 0
			
			mo.slipstreamsoundtimer = false
		end
	else
		mo.slipstreamcharge = max($-1, 0)
	end

	if mo.slipstreamtimer or alwayson then
		clampedkartspeed = Clamp(p.kartspeed, 1, 9)
		slipstreamboost = Rescale(clampedkartspeed*FRACUNIT, 9*FRACUNIT, FRACUNIT, minspeedboostpercent, maxspeedboostpercent)
		slipstreamaccel = accelboostpercent

		if colorized and p.skincolor then
			skincolor = p.skincolor
		else 
			skincolor = SKINCOLOR_WHITE
		end
		
		SpawnFastLines(mo, FRACUNIT, skincolor)
		
		if mo.BLonground then
			SSDust(mo)
		end
	end

	slipstreamboost = min((mo.lastssboost or 0) + $/25, $) -- smoothing, takes some tics to reach full effect
	SetBoost(p, boosttype, slipstreamboost, slipstreamaccel, nil, true, true, false)
	mo.slipstreamtimer = max(($ or 0)-1, 0)
	mo.lastssboost = slipstreamboost
end, MT_PLAYER)

--addHook("ThinkFrame", do
--	AngleFinder()
--end)

addHook("MapLoad", hostmodload)
addHook("NetVars", hostmodload) -- Joiners don't run MapLoad