
if not BLib then return end
local SetBoost = BLib.SetBoost

local prefix = "blib_"
local cv_slopeboost = CV_RegisterVar({
    name = prefix.."slopeboost",
    defaultvalue = "On",
    flags = CV_NETVAR|CV_CALL,
    PossibleValue = CV_OnOff,
	func = onlyAdjustableOnLevelChange
})

local slopeboost
addHook("MapChange", function(gamemap)
    slopeboost = cv_slopeboost.value 
end)
addHook("NetVars", function(net)
    slopeboost = net($) 
end)

local function Rescale(value, oldmin, oldmax, newmin, newmax)
	return newmin +  FixedMul(FixedDiv( value-oldmin , oldmax-oldmin), newmax-newmin)
end

local function Clamp(value, newmin ,newmax)
	return min(newmax,max(newmin,value))
end

local FRACUNIT = FRACUNIT
local ANGLE_180 = ANGLE_180
local ANGLE_90 = ANGLE_90
local ANG1 = ANG1
local MFE_VERTICALFLIP = MFE_VERTICALFLIP
local BT_BRAKE = BT_BRAKE
local k_drift = k_drift
local k_offroad = k_offroad
local k_invincibilitytimer = k_invincibilitytimer
local k_hyudorotimer = k_hyudorotimer

local FixedMul = FixedMul
local FixedDiv = FixedDiv
local min = min
local max = max
local FixedAngle = FixedAngle
local P_ReturnThrustX = P_ReturnThrustX
local P_ReturnThrustY = P_ReturnThrustY
local S_SoundPlaying = S_SoundPlaying
local P_InstaThrust = P_InstaThrust
local R_PointToAngle2 = R_PointToAngle2
local FixedHypot = FixedHypot

local maxboost = FRACUNIT*300/100 -- Rarely ever get a steeper angle than 45, so about half or less is what you commonly see 
local boostcap = maxboost/3 -- the actual cap is here, essentially a 45 degree will give you the max boost, but steeper slopes will get there faster
local maxaccel = FRACUNIT*3
local driftaccelbonus = FRACUNIT*4
local basedecayrate = FRACUNIT/270

local boosttype = "slope"
local boost, accel = 0, 0
local slope, angletodownhill, flip, anglemult, slopemult, weightmult, finalmult, travelangle, lastslopeboost, player, brakemodifier, pks
addHook("MobjThinker", function(mo)
	if not slopeboost then return end
	player = mo.player
	pks = player.kartstuff
	slope = mo.standingslope
	lastslopeboost = mo.lastslopeboost or 0
	
	if (player.cmd.buttons & BT_BRAKE) or pks[k_offroad] then brakemodifier = FRACUNIT*1/100 else brakemodifier = 0 end
	
	if not slope or pks[k_offroad] and not pks[k_hyudorotimer] and not pks[k_invincibilitytimer] and not player.naturalpanelsneakertimer and not player.naturalitemsneakertimer then
		boost = max(lastslopeboost - (basedecayrate + brakemodifier + lastslopeboost/60), 0) -- no slope, but you lose slope boost more slowly than other boosts
		SetBoost(player, boosttype, boost, 0, nil, true, false, false)
		mo.lastslopeboost = boost
		mo.lastslopeaccel = 0
		return
	end
	
	flip = mo.eflags & MFE_VERTICALFLIP
	travelangle = R_PointToAngle2(0, 0, mo.momx, mo.momy)
	
	if (slope.zangle < 0 and not flip) or (slope.zangle > 0 and flip) then
		angletodownhill = travelangle - slope.xydirection
	else
		angletodownhill = travelangle - (slope.xydirection+ANGLE_180)
	end
	
	angletodownhill = max(abs(angletodownhill) - ANG1*3, 0) -- The subtraction means you get a little bit of leeway to get full boost
	
	--local uphillangle
	--if slope.zangle >= 0 then uphillangle = slope.xydirection
	--else uphillangle = slope.xydirection+ANGLE_180 end
	
	--print("playerangle: "..mo.angle.." xydirection: "..slope.xydirection.." zangle: "..slope.zangle.." upphillangle: "..uphillangle)
	
	if angletodownhill >= ANGLE_90 then -- You're facing uphill, no boost for you!
		boost = max(lastslopeboost - (basedecayrate + brakemodifier + lastslopeboost/60), 0) -- but you still get the normal diminish rate
		SetBoost(player, boosttype, boost, 0, nil, true, false, false)
		mo.lastslopeboost = boost
		mo.lastslopeaccel = 0
		return
	end

	anglemult = Rescale(ANGLE_90-angletodownhill, 0, ANGLE_90, 0, FRACUNIT)
	slopemult = Rescale(min(abs(slope.zangle)+ANG1*3, ANGLE_90), 0, ANGLE_90, 0, FRACUNIT) -- Added a little artificial angle here, to raise the floor
	--weightmult =  Rescale(Clamp(player.kartweight, 1, 9), 9, 1, FRACUNIT, FRACUNIT*2)
	
	finalmult = FixedMul(anglemult,slopemult)
	boost = min(FixedMul(maxboost, finalmult), boostcap)
	--accel = FixedMul(FixedMul(maxaccel, weightmult), finalmult)
	accel = FixedMul(maxaccel, finalmult)
	
	if pks[k_drift] then
		accel = $ + FRACUNIT/2 + FixedMul(driftaccelbonus, finalmult) -- Without much accel bonus, drifting just kinda slows you down a lot and
	end													 			-- makes slopes you drift down feel not great, so we give bonus accel when drifting
	
	--print("anglemult: "..anglemult.." slopemult: "..slopemult)	

	if player.speed < FixedMul(FRACUNIT*5, mapobjectscale) or player.cmd.buttons & BT_BRAKE then -- No speedboost if hardly moving, but you still get the accel boost, which allows the bonus to ramp up
		boost = max(lastslopeboost - (basedecayrate + brakemodifier + lastslopeboost/60), 0) 
	elseif boost >= lastslopeboost then
		boost = min(boost, lastslopeboost+finalmult*5/100) -- Smoothing, it takes some tics to get the full benefit of a slope
	else
		boost = max(lastslopeboost - (basedecayrate + brakemodifier + lastslopeboost/60), boost)
	end
	
	SetBoost(mo.player, boosttype, boost, min(accel, (mo.lastslopeaccel or 0)+(accel/8)), nil, true, false, false)
	
	mo.lastslopeboost = boost
	mo.lastslopeaccel = accel
end, MT_PLAYER)


addHook("MapLoad", hostmodload)
addHook("NetVars", hostmodload) -- Joiners don't run MapLoad

/*
local pmo
hud.add(function(v, p, c)
    if p.spectator or not p.mo.standingslope then return end
	pmo = p.mo 
	v.drawString(12,95,  " s xydir: " .. pmo.standingslope.xydirection,V_SNAPTOLEFT,"left")
	v.drawString(12,105, "s zangle: " .. pmo.standingslope.zangle,V_SNAPTOLEFT,"left")
	v.drawString(12,115, "trav ang: " .. (travelangle or "0"),V_SNAPTOLEFT,"left")
	v.drawString(12,125, "ang down: " .. (angletodownhill or "0"),V_SNAPTOLEFT,"left")
    v.drawString(12,135, "spd: " .. FixedMul(p.kartstuff[k_speedboost], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
	v.drawString(12,145, "acc: " .. FixedMul(p.kartstuff[k_accelboost], 100*FRACUNIT)>>FRACBITS,V_SNAPTOLEFT,"left")
end)
*/